local GridStatus = Grid:GetModule("GridStatus");
local GridRoster = Grid:GetModule("GridRoster");
local SmartHealing = GridStatus:GetModule("GridStatusSmartHealing");
local MapData = LibStub("LibMapData-1.0");

local GridStatusSmartHealing_ChiBurst = SmartHealing:NewModule("GridStatusSmartHealing_ChiBurst", "AceEvent-3.0");

GridStatusSmartHealing_ChiBurst.defaultDB =
{
    enable = false,
    color = { r = 1, g = 0, b = 1, a = 0.5 },
    priority = 50,
    color2 = { r = 1, g = 1, b = 0, a = 1 },
    --
    minMissingHealth = 85000, -- TODO change TODO account for DR
    chiBurstWidth = 4, -- square yards
    frameX = 0,
    frameY = 0,
    maxHeal = 32800,
    maxRenew = 22000, -- TODO update with REAL data.
    framesLocked = false,
    showOutOfRange = true, -- TODO make this work.
    displayFrame = false,
    spinningCraneKick = 10 ^ 2, -- Tooltip says 8, testing shows 10
    spinningCraneKickHeal = 6000, -- TODO update with real data
    --chiBurstBloomSize = 5^2, -- square yards
    --rangeFromTarget = 10,
    --sortMethod = "missingHealth",
    --minTargets = 3,
    --maxStatuses = 15,
    --centerTextOption = "bestTarget",
};

-- BaseLine
local ChiBase = (432 + 1295) / 2; --http://www.wowhead.com/spell=130654

-- SpellMod

local Chi_Mod = 0.789; -- http://www.wowhead.com/spell=130654#comments

-- Line Class...
local Line = {};
Line.__index = Line;

function Line.create(x1, y1, x2, y2)
	local line = {};
	setmetatable(line, Line);
	line.x1 = x1;
	line.y1 = y1;

	line.x2 = x2;
	line.y2 = y2;
	return line;
end

function Line:has(x, y)
	return (x >= self.x1 and x <= self.x2) or (x <= self.x1 and x >= self.x2);
end
function Line:rise()
	return self.y2 - self.y1;
end
function Line:run()
	return self.x2 - self.x1;
end
function Line:slope()
	return self:rise() / self:run();
end
function Line:length()
	return math.sqrt((self:rise() * self:rise()) + (self:run() * self:run()));
end
function Line:intercept()
	return self.y1 - (self:slope() * self.x1);
end
function Line:intersect(line)
	local x = (line:intercept() + (self:intercept() * -1)) / (self:slope() + (line:slope() * -1));
	local y = (self:slope() * x) + self:intercept();

	return  x, y;
end

local options =
{
	["framesLocked"] =
	{
		order = 100,
		type = "toggle",
		name = "Frames Locked?",
		desc = "",
		get = function()
			return GridStatusSmartHealing_ChiBurst.db.profile.framesLocked;
		end,
		set = function(info, v)
			GridStatusSmartHealing_ChiBurst:FrameLock(v);
		end
	},
	["minMissingHealth"] =
    {
        order = 110,
        type = "range",
        name = "Min Total Missing Health",
        desc = "Threshold for a targets (total)heals to have, before it shows.",
        min = 0,
        max = 200000,
        step = 1,
        width = "full",
        get = function()
            return GridStatusSmartHealing_ChiBurst.db.profile.minMissingHealth;
        end,
        set = function(info, v)
            GridStatusSmartHealing_ChiBurst.db.profile.minMissingHealth = v;
            GridStatusSmartHealing_ChiBurst:ClearAll();
            GridStatusSmartHealing_ChiBurst:RefreshAll();
        end
    },
    ["color2"] = 
    {
        order = 21,
        type = "color",
        name = "Color x2",
        desc = string.format("Color for Best Chi Burst Target", desc),
        hasAlpha = true,
        get = function()
            local color = GridStatusSmartHealing_ChiBurst.db.profile.color2;
            return color.r, color.g, color.b, color.a;
        end,
        set = function(info, r, g, b, a)
            local color = GridStatusSmartHealing_ChiBurst.db.profile.color2;
            color.r = r;
            color.g = g;
            color.b = b;
            color.a = a or 1;
        end,
    },
    ["displayFrame"] =
    {
		order = 120,
    	type = "toggle",
    	name ="Enable Monk Bar",
    	desc = "Display Renewing Mist vs Chi Burst vs spinningCraneKick",
    	get = function()
    		return GridStatusSmartHealing_ChiBurst.db.profile.displayFrame;
    	end,
    	set = function(info, v)
    		GridStatusSmartHealing_ChiBurst.db.profile.displayFrame = v;
    		GridStatusSmartHealing_ChiBurst:ClearAll();
            GridStatusSmartHealing_ChiBurst:RefreshAll();
    	end,
	},

	--[[
    ["test"] =
    {
    	order = 120,
    	type = "toggle",
    	name ="test",
    	desc = "Test",
    	get = function()
    	end,
    	set = function(info, v)
    		GridStatusSmartHealing_ChiBurst.db.profile.frameX = 0;
    		GridStatusSmartHealing_ChiBurst.db.profile.frameY = 0;
			GridStatusSmartHealing_ChiBurst:FramePos();
			print("Reset Frame");
			print(GridStatusSmartHealing_ChiBurst.db.profile.frameX, GridStatusSmartHealing_ChiBurst.db.profile.frameY);
			print(GridStatusSmartHealing_ChiBurst.frame_chi:GetLeft(), GridStatusSmartHealing_ChiBurst.frame_chi:GetTop());
    	end,
	}
	]]
};

local thisStatus = "alert_gssh_chiburst";
local Settings;

function GridStatusSmartHealing_ChiBurst:PostInitialize()
	SmartHealing:RegisterStatus(self, thisStatus, "SH: Monk - Chi Burst", options, false);

	Settings = self.db.profile;

	-- TODO add more setting for frame.
	self:RefreshAll();
	--GridStatus:RegisterMessage("Grid_StatusGained", function(...) self:Grid_StatusGained2(...); end);
end

function GridStatusSmartHealing_ChiBurst:Grid_StatusGained2(message, guid, status, priority, range, color, text, value, maxValue, texture, start, duration, stack, texCoords)
	if (status ~= "buff_RenewingMist") then
		return;
	end

	--local unit = GridStatusSmartHealing:GetUnitInfo(guid);

	self:UpdateRenewing();
end

function GridStatusSmartHealing_ChiBurst:UpdateRenewing()
	local totalMissingHealth = 0;

	local ret = GridStatus:CachedStatusIterator("buff_RenewingMist");
	local i = 0;
	for guid, status, table in ret do
		i = i +1;
		local unit = GridStatusSmartHealing:GetUnitInfo(guid);
		if (unit) then
			if (unit.missingHealth > Settings.maxRenew) then
				totalMissingHealth = Settings.maxRenew + totalMissingHealth;
			else
				totalMissingHealth = unit.missingHealth + totalMissingHealth;
			end
		end
	end
	--print("Targets = ", i);
	--print("Total Missing Health", totalMissingHealth);
	self:SetRenewTargets(i);
	self:SetRenewEstimated(totalMissingHealth);
end

function GridStatusSmartHealing_ChiBurst:UpdateSpinningCraneKick()
	local player = GridRoster:GetGUIDByName(UnitName("player"));
	player = SmartHealing:GetUnitInfo(player);

	local totalMissingHealth = 0;
	local totalCount = -1;

	for guid, unitId in GridRoster:IterateRoster() do
		local unit = SmartHealing:GetUnitInfo(guid);

		if (unit and unit.guid and  SmartHealing:IsValidHealTarget(unit.unitId)) then
			local distance = SmartHealing:Distance(player, unit);
			if (distance ~= nil and distance <= Settings.spinningCraneKick) then
				totalCount = totalCount + 1;
				if (unit.missingHealth > Settings.spinningCraneKickHeal) then
					totalMissingHealth = totalMissingHealth + Settings.spinningCraneKickHeal;
				else
					totalMissingHealth = totalMissingHealth + unit.missingHealth;
				end
			end
		end
	end

	self:SetSpinningKickTargets(totalCount);
	self:SetSpinningKickEstimated(totalMissingHealth);
end

function GridStatusSmartHealing_ChiBurst:CreateChiFrame()
	local frame_chi = CreateFrame("Button", "GSSH_ChiBurst_Button");
	--frame_chi:SetFrameLevel(frame:GetFrameLevel()+1);
	frame_chi:SetWidth(20);
	frame_chi:SetHeight(20);
	frame_chi:ClearAllPoints();
	frame_chi:SetPoint("BOTTOMLEFT", 0, 0);

	local texture = frame_chi:CreateTexture(nil, "PARENT");
	texture:SetAllPoints(frame_chi);
	--texture:SetTexture("Interface\\Icons\\ability_monk_powerstrikes");
	texture:SetTexture("Interface\\Icons\\Spell_Arcane_ArcaneTorrent");

	self.frame_chi = frame_chi;

	local text = frame_chi:CreateFontString("GSSH_ChiBurstText", "OVERLAY", "GameFontHighlightSmall");
	text:SetFont("Fonts\\FRIZQT__.TTF", 11, "OUTLINE");
	text:SetPoint("CENTER", frame_chi, "CENTER", 0, 0);
	text:SetText("");

	self.frame_chi_text = text;

	local text = frame_chi:CreateFontString("GSSH_ChiBurstTextHealing", "OVERLAY", "GameFontHighlightSmall");
	text:SetFont("Fonts\\FRIZQT__.TTF", 11, "OUTLINE");
	text:SetPoint("LEFT", frame_chi, "CENTER", 20, 0);
	text:SetText("");

	self.frame_chi_heal = text;

	-- Add Renewing Mist

	local frame = CreateFrame("Frame", "GSSH_ChiBurst_RenewingMist", frame_chi);
	frame:SetWidth(20);
	frame:SetHeight(20);
	frame:SetPoint("CENTER", frame_chi, "CENTER", 0, -20);

	self.frame_renewing = frame;

	local texture = frame:CreateTexture(nil, "PARENT");
	texture:SetAllPoints(frame);
	texture:SetTexture("Interface\\Icons\\ability_monk_renewingmists");

	local text = frame:CreateFontString("GSSH_ChiBurst_Renew", "OVERLAY", "GameFontHighlightSmall");
	text:SetFont("Fonts\\FRIZQT__.TTF", 11, "OUTLINE");
	text:SetPoint("CENTER", frame, "CENTER", 0, 0);
	text:SetText("");

	self.renewing_text = text;

	local text = frame:CreateFontString("GSSH_ChiBurst_RenewHeal", "OVERLAY", "GameFontHighlightSmall");
	text:SetFont("Fonts\\FRIZQT__.TTF", 11, "OUTLINE");
	text:SetPoint("LEFT", frame, "CENTER", 20, 0);
	text:SetText("");

	self.renewing_heal = text;

	-- Add Spinning Crane Kick

	local frame = CreateFrame("Frame", "GSSH_ChiBurst_SpinningCrane", frame_chi);
	frame:SetWidth(20);
	frame:SetHeight(20);
	frame:SetPoint("CENTER", frame_chi, "CENTER", 0, -40);

	self.frame_spinningkick = frame;

	local texture = frame:CreateTexture(nil, "PARENT");
	texture:SetAllPoints(frame);
	texture:SetTexture("Interface\\Icons\\ability_monk_cranekick_new");

	local text = frame:CreateFontString("GSSH_ChiBurstText_cranekick", "OVERLAY", "GameFontHighlightSmall");
	text:SetFont("Fonts\\FRIZQT__.TTF", 11, "OUTLINE");
	text:SetPoint("CENTER", frame, "CENTER", 0, 0);
	text:SetText("");

	self.spinningkick_text = text;

	local text = frame:CreateFontString("GSSH_ChiBurstText_cranekickheal", "OVERLAY", "GameFontHighlightSmall");
	text:SetFont("Fonts\\FRIZQT__.TTF", 11, "OUTLINE");
	text:SetPoint("LEFT", frame, "CENTER", 20, 0);
	text:SetText("");

	self.spinningkick_heal = text;

	self:FramePos();

	self:FrameLock(Settings.framesLocked);
end

function GridStatusSmartHealing_ChiBurst:FramePos()
	-- TODO Fix pos
	self.frame_chi:ClearAllPoints();
	self.frame_chi:SetPoint("BOTTOMLEFT", Settings.frameX, Settings.frameY);
	--self.frame_chi:ClearAllPoints();
	--self.frame_chi:SetPoint("TOPLEFT", 0, 0);

	self.frame_chi:RegisterForDrag("LeftButton");
	self.frame_chi:SetScript("OnDragStart", self.frame_chi.StartMoving);
	self.frame_chi:SetScript("OnDragStop", self.frame_chi.StopMovingOrSizing);
end

function GridStatusSmartHealing_ChiBurst:FrameLock(lock)
	--print(Settings.frameX, Settings.frameY);
	Settings.framesLocked = lock;

	-- No Frames Exist. end.
	if (not self.frame_chi) then
		return;
	end

	if (lock) then
		Settings.frameX = self.frame_chi:GetLeft();
		Settings.frameY = self.frame_chi:GetBottom();
	end

	--print(Settings.frameX, Settings.frameY);

	self.frame_chi:SetMovable(not lock);
	self.frame_chi:EnableMouse(not lock);
end

function GridStatusSmartHealing_ChiBurst:HideChiFrame()
	if (self.frame_chi and self.frame_chi:IsVisible()) then
		self.frame_chi:Hide();
		--self.frame_renewing:Hide();
		return true;
	end

	return false;
end

function GridStatusSmartHealing_ChiBurst:ShowChiFrame()
	if ((not self.frame_chi) or self.frame_chi:IsVisible()) then
		return false;
	end

	self.frame_chi:Show();
	--self.frame_renewing:Show();
	return true;
end

function GridStatusSmartHealing_ChiBurst:SetChiTargets(target_num)
	if (not Settings.displayFrame) then return; end
	self.frame_chi_text:SetText(target_num);
end

function GridStatusSmartHealing_ChiBurst:SetChiEstimated(estimatedHeal)
	if (not Settings.displayFrame) then return; end
	self.frame_chi_heal:SetText(estimatedHeal);
end

function GridStatusSmartHealing_ChiBurst:SetRenewTargets(target_num)
	if (not Settings.displayFrame) then return; end
	self.renewing_text:SetText(target_num);
end

function GridStatusSmartHealing_ChiBurst:SetRenewEstimated(estimatedHeal)
	if (not Settings.displayFrame) then return; end
	self.renewing_heal:SetText(estimatedHeal);
end

function GridStatusSmartHealing_ChiBurst:SetSpinningKickTargets(target_num)
	if (not Settings.displayFrame) then return; end
	self.spinningkick_text:SetText(target_num);
end

function GridStatusSmartHealing_ChiBurst:SetSpinningKickEstimated(estimatedHeal)
	if (not Settings.displayFrame) then return; end
	self.spinningkick_heal:SetText(estimatedHeal);
end

function GridStatusSmartHealing_ChiBurst:OnStatusEnable(status)
	SmartHealing:DoEnable();

	self:RefreshAll();
end

function GridStatusSmartHealing_ChiBurst:OnStatusDisable(status)
	SmartHealing:DoDisable();
	---
	--GridStatus:UnregisterMessage("Grid_StatusGained");

	self:ClearAll();
end

function GridStatusSmartHealing_ChiBurst:Update()
	self:ClearAll();
	self:UpdateRenewing();
	self:UpdateSpinningCraneKick();
	-- all players missing health.
	local destinations = {};

	local player = GridRoster:GetGUIDByName(UnitName("player"));
	player = SmartHealing:GetUnitInfo(player);
	player.greenLength = NaN;

	local finalTargets = {};

	for guid, unitId in GridRoster:IterateRoster() do
		local destination = SmartHealing:GetUnitInfo(guid);
		if (destination ~= nil and unitId and SmartHealing:IsValidHealTarget(unitId)) then
			destination.targets = {};
			destination.greenLength = NaN;
			destination.totalHealth = 0;
			destination.estimatedHeal = 0;

			local PlayerToDestination = Line.create(player.x, player.y, destination.x, destination.y);

			for guid2, unitId2 in GridRoster:IterateRoster() do
				-- cant hit self twice..
				local hit = SmartHealing:GetUnitInfo(guid2);

				if (hit ~= nil and guid ~= guid2 and guid2 ~= player.guid) then
					-- Fake GreenLine
					local GreenLine = Line.create(hit.x, hit.y, hit.x + PlayerToDestination:rise(), hit.y + (PlayerToDestination:run() * -1));

					local x, y = GreenLine:intersect(PlayerToDestination);

					if (PlayerToDestination:has(x, y)) then
						-- Fix GreenLine
						GreenLine.x2 = x;
						GreenLine.y2 = y;

						local greenLength = GreenLine:length();

						if (greenLength <= Settings.chiBurstWidth) then
							hit.greenLength = greenLength;
							destination.totalHealth = destination.totalHealth + hit.missingHealth;
							if (hit.missingHealth > Settings.maxHeal) then
								destination.estimatedHeal = destination.estimatedHeal + Settings.maxHeal;
							else
								destination.estimatedHeal = destination.estimatedHeal + hit.missingHealth;
							end
							
							table.insert(destination.targets, hit);
						end
					end
				end
			end
			-- Remove 0 hits
			if (#destination.targets > 0 and destination.estimatedHeal >= Settings.minMissingHealth) then
				table.insert(finalTargets, destination);
			end
			--[[
			self:Debug(UnitName(GridRoster:GetUnitidByGUID(destination.guid)), #destination.targets);
			if (#destination.targets > 0) then
				for index, target in pairs(destination.targets) do
					self:Debug("    --", UnitName(GridRoster:GetUnitidByGUID(target.guid)), target.greenLength);
				end
			end
			]]--
		end
	end

	table.sort(finalTargets, function(a, b)
		return a.estimatedHeal > b.estimatedHeal;
	end);

	-- Clear Text
	self:SetChiTargets("");
	self:SetChiEstimated("");

	for i, target in pairs(finalTargets) do
		local text = string.format("#%s", #target.targets);
		local color = Settings.color;
		if (i == 1) then
			-- grab best color
			color = Settings.color2;
			self:SetChiTargets(text);
			self:SetChiEstimated(target.estimatedHeal);
		end
		self.core:SendStatusGained(target.guid, thisStatus, Settings.priority, nil, color, string.format("%s", text), 1, nil, nil);
	end
end

function GridStatusSmartHealing_ChiBurst:RefreshAll()
	if (Settings.displayFrame) then
		if (not self:ShowChiFrame()) then
			self:CreateChiFrame();
		end
	else
		self:HideChiFrame();
	end
end

function GridStatusSmartHealing_ChiBurst:ClearAll()
    self.core:SendStatusLostAllUnits(thisStatus);
end